﻿/************************************************************
 *	File:		CCRServer.cs
 *
 * 	Author:		Samuel Joski
 *	
 *	Contents:	Class GLOBALS
 *				Class CCRServer
 * 	
 * 	Purpose:	The GLOBALS class is used to represent the
 * 				meaning of railroad resource various statuses
 * 				and directions.
 * 				
 * 				The CCRServer class is responsible for handling
 * 				reading and writing to the files for trains,
 * 				turnouts, tracks, and AIUs, as well as for
 * 				communicating any changes in the file to the
 * 				appropriate railroad resource objects. It is
 * 				also responsible for updating the Form to
 * 				display the information about the state of the
 * 				railroad.		
 * *********************************************************/
using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Timers;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Windows.Forms;

namespace CCRServer
{
	public partial class CCRServer : Form
	{
		//Methods to call the DCC
		public static CommandDCC CallDCC;

		//Railroad Data Structures
		private AIU[] AIUs;
		private Sensor[] sensors;
		private Track[] tracks;
		private Train[] trains;
		private Turnout[] turnouts;

		//Timers
		private System.Timers.Timer fileReader;
		private System.Timers.Timer AIUPoller;

		//Dynamic Form Elements
		private Label[] lblTracks;
		private TextBox[] tbTracks;

		public CCRServer(string com_port)
		{
			try
			{
				CallDCC = new CommandDCC(com_port);
			}
			catch(Exception)	//Any exception indicated the port is unavailable
			{
				showNotification(GLOBALS.FAILED_PORT);
			}

			//Read in the files. IOException if files fail
			try
			{
				setupAIUsAndSensors();
				setupTurnouts();
				setupTracks();
				setupTrains();
			}
			catch (IOException)
			{
				return;
			}

			//Setup fileReader timer
			fileReader = new System.Timers.Timer();
			fileReader.Elapsed += new ElapsedEventHandler(readFilesEvent);
			fileReader.Interval = 100;

			//Setup AIUPoller Timer
			AIUPoller = new System.Timers.Timer();
			AIUPoller.Elapsed += new ElapsedEventHandler(pollAIUEvent);
			AIUPoller.Interval = 100;
			
			//Initialize the Form elements
			InitializeComponent();
			setupForm();
			updateForm();

			return;
		}

		//CCRServer Form Functions
		private void onShow(object sender, EventArgs e)
		{
			//Start the Timers
			fileReader.Enabled = true;
			AIUPoller.Enabled = true;
		}
		private void setupForm()
		{
			/**************************************************************
			 *	Function Name:		setupForm
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function creates the Labels and TextBoxes
			 *				for each Track section to appear on the form. 
			 *				The sizing of the elements are pulled from the
			 *				GLOBALS class.
			 *				
			 *				The Labels show the track that the adjacent
			 *				TextBox will display information for. The
			 *				TextBoxes are all initialized to display FREE,
			 *				but will display when a train is either ON or
			 *				OWNS the section.
			 * ************************************************************/
			lblTracks = new Label[tracks.Length];
			tbTracks = new TextBox[tracks.Length];

			for (int i = 0; i < tracks.Length; i++)
			{
				//Create the Track's Label
				lblTracks[i] = new Label();
				lblTracks[i].Text = ("Track " + i.ToString() + ":");
				lblTracks[i].Height = GLOBALS.lblHeight;
				lblTracks[i].Width = GLOBALS.lblWidth;

				//Create the Track's TextBox
				tbTracks[i] = new TextBox();
				tbTracks[i].Text = GLOBALS.FREE;
				tbTracks[i].ReadOnly = true;
				tbTracks[i].Height = GLOBALS.tbHeight;
				tbTracks[i].Width = GLOBALS.tbWidth;

				//Place the Label based on the number of them
				int x = GLOBALS.xBuffer + ((GLOBALS.lblWidth + GLOBALS.tbWidth + GLOBALS.xOffset) * (i / 8));
				int y = GLOBALS.yBuffer + ((GLOBALS.lblHeight + GLOBALS.yOffset) * (i % 8));
				lblTracks[i].Location = new Point(x, y);

				//Place TextBox with the same coordinates as the Label, just offset by the width of the Label
				x += GLOBALS.lblWidth;
				tbTracks[i].Location = new Point(x, y);

				//Add the Label and TextBox to the GroupBox gbTracks
				gbTracks.Controls.Add(lblTracks[i]);
				gbTracks.Controls.Add(tbTracks[i]);
			}
			//Size the GroupBox based on the elements in it
			gbTracks.Height = (GLOBALS.yBuffer + ((tracks.Length % 8) + 1) * (GLOBALS.lblHeight + GLOBALS.yOffset));
			gbTracks.Width = (GLOBALS.xBuffer + ((tracks.Length / 8) + 1) * ((GLOBALS.lblWidth + GLOBALS.tbWidth) + GLOBALS.xOffset));
		}
		private void updateForm()
		{
			/**************************************************************
			 *	Function Name:		updateForm
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function updates the TextBoxes for each
			 *				Track section to appear on the form.
			 *				
			 *				If a track is held, all of the trains are
			 *				checked. If the train is AT the location being
			 *				checked, GLOBALS.AT is used for the TextBox
			 *				text. If the train OWNS the locations being
			 *				checked, GLOBALS.OWNS is used for the text.
			 *				
			 *				If no train is on the location, GLOBALS.FREE is
			 *				used for the text.
			 *				
			 *				This function also adjusts the train displays
			 *				to correctly show if a train is occupied, or if
			 *				it is free to be used.				
			 * ************************************************************/

			//Loop through all Track sections
			for (int i = 0; i < tracks.Length; i++)
			{
				//If the section is held by a train
				if(tracks[i].isHeld())
				{
					//Loop through all the trains to see which one holds the track
					for (int j = 0; j < trains.Length; j++)
					{
						//If the index is not null
						if (trains[j] != null)
						{
							//If the train location matches
							if (trains[j].Location() == tracks[i])
							{
								//Set the track's TextBox to indicate that the train is AT the section
								tbTracks[i].Text = GLOBALS.AT.Replace('x', Convert.ToChar(trains[j].Address().ToString()));
							}

							//Make sure the train's Next is not null before calling its methods
							else if (trains[j].Next() != null)
							{
								//If the train Next location matches
								if (trains[j].Next() == tracks[i])
								{
									//Set the track's TextBox to indicate that the train OWNS the secion
									tbTracks[i].Text = GLOBALS.OWNS.Replace('x', Convert.ToChar(trains[j].Address().ToString()));
								}
							}
						}
					}
				}

				//If the track was free, then change the text to indicate it
				else
				{
					//Set the track's TextBox to indicate that it is FREE
					tbTracks[i].Text = GLOBALS.FREE;
				}
			}

			//Update train displays by looping through train array
			for (int i = 0; i < trains.Length; i++)
			{
				//If valid index
				if (trains[i] != null)
				{
					//If the train is occupied
					if (trains[i].isOccupied())
					{
						//Check if the train is not already in the ListBox of occupied trains before adding
						if (!lbOccupiedTrains.Items.Contains("Engine " + trains[i].Address()))
						{
							lbOccupiedTrains.Items.Add("Engine " + trains[i].Address());
						}

						//Check if the train is in the ListBox of available trains before removing it
						if (lbAvailableTrains.Items.Contains("Engine " + trains[i].Address()))
						{
							lbAvailableTrains.Items.Remove("Engine " + trains[i].Address());
						}
					}

					//Else the train is not occupied
					else
					{
						//Check if the train is not already in the ListBox of available trains before adding
						if (!lbAvailableTrains.Items.Contains("Engine " + trains[i].Address()))
						{
							lbAvailableTrains.Items.Add("Engine " + trains[i].Address());
						}

						//Check if the train is in the ListBox of occupied trains before removing it
						if (lbOccupiedTrains.Items.Contains("Engine " + trains[i].Address()))
						{
							lbOccupiedTrains.Items.Remove("Engine " + trains[i].Address());
						}
					}
				}
			}
		}
		private void onDisconnectClick(object sender, EventArgs e)
		{
			Close();
		}
		private void onExitClick(object sender, EventArgs e)
		{
			// Confirm user wants to close
			switch (MessageBox.Show(this, "Are you sure you want to disconnect?", "Disconnect", MessageBoxButtons.YesNo))
			{
				case DialogResult.No:
					break;
				default:
					Application.Exit();
					break;
			}
		}
		private void onAddTrainClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onAddTrainClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is called when Edit -> Add Train
			 *				is clicked on the CCRServer form. An AddTrain
			 *				form is created and displayed to prompt the
			 *				user for a new train to be added to the Server.
			 *				
			 *				The user must enter:
			 *					* An Address
			 *					* A Location
			 *					* An Orientation
			 *					
			 *				If the address or location is already in use,
			 *				a notification is displayed to the user, and the
			 *				train is NOT added to the Server.
			 *				
			 *				If all information is valid, the train is added,
			 *				WriteTrain is called, and the CCRServer form is
			 *				updated for the new train.
			 * ************************************************************/

			AddTrain addtrain = new AddTrain();
			var result = addtrain.ShowDialog();

			//If the form clicked the Confirm Button
			if (result == DialogResult.OK)
			{
				//Loop through the train array looking for an available index
				for (int i = 0; i < trains.Length; i++)
				{
					//If the index is available
					if (trains[i] == null)
					{
						//Try to parse the train location. Display error if invalid location
						Track tempTrack;
						try
						{
							tempTrack = tracks[Convert.ToInt32(addtrain.TrainLocation)];
						}

						//Display message and return if invalid section
						catch (IndexOutOfRangeException)
						{
							showNotification(GLOBALS.FAILED_TRAIN_ADD_LOCATION_INVALID);
							return;
						}

						/*Make sure the entered Track location is not currently held by 
							another train.*/
						if (tempTrack.isHeld())
						{
							showNotification(GLOBALS.FAILED_TRAIN_ADD_LOCATION_TAKEN);
							return;
						}

						/*Make sure the entered Train address is not currently in use
							by another train.*/
						for (int j = 0; j < trains.Length; j++)
						{
							if (trains[j] != null)
							{
								if (trains[j].Address() == addtrain.TrainAddress)
								{
									showNotification(GLOBALS.FAILED_TRAIN_ADD_ADDRESS);
									return;
								}
							}
						}

						System.Threading.Monitor.Enter(trains);
						System.Threading.Monitor.Enter(tracks);

						//Add the train to the train array
						bool added = false;
						try
						{
							trains[i] = new Train(
								i,
								addtrain.TrainAddress,
								tempTrack,
								addtrain.TrainOrientation
							);
							added = true;
						}

						//Catch that the constructor was unable to request the location
						catch (ArgumentException)
						{
							showNotification(GLOBALS.FAILED_TRAIN_ADD_LOCATION_TAKEN);
						}

						if (added)
						{
							//Add the train to the Form display and logic
							writeTrain(trains[i]);
							lbAvailableTrains.Items.Add("Engine " + trains[i].Address());
						}

						System.Threading.Monitor.Exit(tracks);
						System.Threading.Monitor.Exit(trains);
						return;
					}
				}
			}
		}
		private void onRemoveTrainClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onRemoveTrainClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is called when Edit -> Remove Train
			 *				is clicked on the CCRServer form. A RemoveTrain
			 *				form is created and displayed to prompt the
			 *				user for a train to be deleted from the Server.
			 *				
			 *				The user must enter:
			 *					* An Address
			 *					
			 *				If the address is already in use, a notification
			 *				is displayed to the user, and the train is NOT
			 *				deleted from the Server.
			 *				
			 *				If all information is valid, the train is deleted,
			 *				WriteTrain is called, and the CCRServer form is
			 *				updated from the deleted train.
			 * ************************************************************/

			DeleteTrain deletetrain = new DeleteTrain();
			var result = deletetrain.ShowDialog();

			//If the form clicked the Confirm Button
			if (result == DialogResult.OK)
			{
				bool found = false;		//Flag to see if the entered address is valid

				//Loop through the train array looking for the address of the train
				for (int i = 0; i < trains.Length; i++)
				{
					//If the index is valid
					if (trains[i] != null)
					{
						//See if its address matches the one requesting to be deleted
						if(trains[i].Address() == deletetrain.TrainAddress)
						{
							System.Threading.Monitor.Enter(trains);
							System.Threading.Monitor.Enter(tracks);
							
							//Remove the train to the Form display and logic
							if(trains[i].Delete())
							{
								trains[i] = null;
								writeDeletedTrain(i);
								lbAvailableTrains.Items.Remove("Engine " + deletetrain.TrainAddress);
							}

							//If unable to delete, it is occupied by a user. Display notification
							else
							{
								showNotification(GLOBALS.FAILED_TRAIN_DELETE_OCCUPIED);
							}

							found = true;

							System.Threading.Monitor.Exit(tracks);
							System.Threading.Monitor.Exit(trains);
						}
					}
				}

				//If the entered address was not found, it was invalid. Display notification
				if(!found)
				{
					showNotification(GLOBALS.FAILED_TRAIN_DELETE_INVALID);
				}
			}
		}
		private void showNotification(int message)
		{
			switch (message)
			{
				case GLOBALS.STOPPED:
					MessageBox.Show("An unexpected sensor was tripped. Either a train had missed a previous sensor, or something went wrong with a sensor. To resume, please place all trains according to the 'Track Information' display.", "UNSAFE CONDITIONS");
					break;
				case GLOBALS.FAILED_TRAIN_ADD_ADDRESS:
					MessageBox.Show("Sorry, but the train was unable to be added. The entered address is already in use.", "Failed to add train");
					break;
				case GLOBALS.FAILED_TRAIN_ADD_LOCATION_INVALID:
					MessageBox.Show("Sorry, but the train was unable to be added. An invalid location was entered.", "Failed to add train");
					break;
				case GLOBALS.FAILED_TRAIN_ADD_LOCATION_TAKEN:
					MessageBox.Show("Sorry, but the train was unable to be added. A currently occupied location taken was entered.", "Failed to add train");
					break;
				case GLOBALS.FAILED_TRAIN_DELETE_OCCUPIED:
					MessageBox.Show("Sorry, but the train was unable to be added. A currently occupied train address was entered.", "Failed to delete train");
					break;
				case GLOBALS.FAILED_TRAIN_DELETE_INVALID:
					MessageBox.Show("Sorry, but the train was unable to be added. An invalid train address was entered.", "Failed to delete train");
					break;
				case GLOBALS.FAILED_PORT:
					MessageBox.Show("Something went wrong with the DCC Communication. Shutting down...", "DCC Communication Error");
					Application.Exit();
					break;
			}
		}
		protected override void OnFormClosing(FormClosingEventArgs e)
		{
			base.OnFormClosing(e);

			if (e.CloseReason == CloseReason.WindowsShutDown)
			{
				return;
			}

			//Disable timers
			AIUPoller.Enabled = false;
			fileReader.Enabled = false;

			//Delete any trains to ensure they stop moving when the application closes
			for (int i = 0; i < trains.Length; i++)
			{
				if (trains[i] != null)
				{
					trains[i].ForceDelete();
					writeDeletedTrain(i);
					trains[i] = null;
				}
			}
			try
			{
				CallDCC.Disconnect();   //Make sure the port closes when the Form closes
			}
			catch { /*Nothing to do*/}
		}
			
		//Timer ElapsedEventHandler Functions
		private void readFilesEvent(object source, ElapsedEventArgs e)
		{
			/**************************************************************
			 *	Function Name:		readFilesEvent
			 *	Input Parameters:	object source, ElapsedEventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is the event triggered by the 
			 *				fileReader timer, which is triggered every 100ms.
			 *				
			 *				The functions readTrains() and readTurnouts() are
			 *				called from this event to update the structures if
			 *				they contain new values, which occurs when CABS
			 *				switch turnouts or send commands for a train.
			 * ************************************************************/

			/*For the whole handler, enter a monitor for calls to the DCC to prevent
				other threads from interfering.*/
			System.Threading.Monitor.Enter(CallDCC);

			//Read the turnouts, entering a monitor so that other threads cannot interfere
			System.Threading.Monitor.Enter(turnouts);
			readTurnouts();
			System.Threading.Monitor.Exit(turnouts);

			//Read the trains, entering a monitor so that other threads cannot interfere
			System.Threading.Monitor.Enter(trains);
			readTrains();
			System.Threading.Monitor.Exit(trains);

			//Update the Form
			Invoke(new MethodInvoker(delegate () {
				updateForm();
			}));

			/*Exit monitor for DCC calls at handler completion.*/
			System.Threading.Monitor.Exit(CallDCC);
		}
		private void pollAIUEvent(object source, ElapsedEventArgs e)
		{
			/**************************************************************
			 *	Function Name:		pollAIUEvent
			 *	Input Parameters:	object source, ElapsedEventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is the event triggered by the 
			 *				AIUPoller timer, which is triggered every 100ms.
			 *				
			 *				The CommandDCC object's function PollAIU is
			 *				called to read from the DCC the states of the
			 *				AIUS to determine if any of the sensors have
			 *				been triggered, and which train hit the sensor.
			 *				
			 *				The trains file and the form are updated to
			 *				reflect any trains the poll found to have moved.
			 * ************************************************************/

			/*For the whole handler, enter a monitor for calls to the DCC to prevent
				other threads from interfering.*/
			System.Threading.Monitor.Enter(CallDCC);

			//Enter a monitor to prevent other AIU poll threads from interfering
			System.Threading.Monitor.Enter(trains);

			//Call poll AIU one by one
			for (int i = 0; i < AIUs.Length; i++)
			{
				//Poll of the AIU will return any trains that moved.
				Train[] movedTrains = null;
				try
				{
					movedTrains = CallDCC.PollAIU(AIUs[i], trains);
				}
				catch (TimeoutException)
				{
					showNotification(GLOBALS.FAILED_PORT);
					return;	//This return should not occur, FAILED_PORT notification closes application
				}
				
				//Write to the trains file any trains that have moved and update the Form
				for (int j = 0; j < movedTrains.Length; j++)
				{
					if (movedTrains[j] != null)
					{
						writeTrain(movedTrains[j]);
						Invoke(new MethodInvoker(delegate ()
						{
							updateForm();
						}));
					}
				}
			}
			System.Threading.Monitor.Exit(trains);

			/*Exit monitor for DCC calls at handler completion.*/
			System.Threading.Monitor.Exit(CallDCC);
		}

		//Setup Data Structures Functions
		private void setupAIUsAndSensors()
		{
			/**************************************************************
			 *	Function Name:		setupAIUs
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function reads from the "aius.txt" file
			 *				to input into the AIUs array all AIUs that
			 *				are present in the file.
			 *				
			 *				The input parameters to the AIU class are:
			 *					* ID			int			0
			 *					* Address		int			1
			 *					* Sensors		Sensor		2+
			 *										
			 *				This function also sets up the Sensor objects
			 *				array by reading how many values are present on
			 *				a line in "aius.txt" past the address element
			 *				(sensors start at the third value in the file).
			 *				
			 *				The input parameters to the Sensor class are:
			 *					* ID			int
			 *					* Status		int(READ FROM DCC)
			 *					
			 *				The status of a sensor is set from reading from
			 *				the COM port for AIUs. The default value from
			 *				the DCC box is set to all sensors as the starting
			 *				value. (Currently unsure of what that value is.)
			 * ************************************************************/

			string[] lines = File.ReadAllLines("aius.txt");
			AIUs = new AIU[lines.Length];

			int numSensors = 0;
			for (int i = 0; i < lines.Length; i++)
			{
				//Split the file line (comma delimmited)
				string[] lineInfo = lines[i].Split(',');

				/*Add the number of sensors present on an AIUs line, excluding
					blank values on the line*/
				for (int j = 2; j < lineInfo.Length; j++)
				{
					if (lineInfo[j] != "")
					{
						numSensors++;
					}
				}
			}

			sensors = new Sensor[numSensors];

			/*Create n AIU objects, where n is the number of lines in the file*/
			for (int i = 0; i < lines.Length; i++)
			{
				//Split the file line (comma delimmited)
				string[] lineInfo = lines[i].Split(',');

				//Create the AIU
				AIUs[i] = new AIU(
					Convert.ToInt32(lineInfo[0]),       //ID
					Convert.ToInt32(lineInfo[1])        //Address
				);

				/*If there are sensors on the AIU line in the file, then
					create a temporary sensor array to give to the AIU object
					(this is different from the array "sensors", in that
					the temporary sensor only holds sensors that are assigned
					to the AIU.*/
				if (lineInfo.Length > 2)
				{
					/*Subtract two, as the first two values in lineInfo are
						the ID and the address of the AIU.*/
					Sensor[] tempSensors = new Sensor[lineInfo.Length - 2];

					/*Create m Sensor objects, where m is the number of values on
						an AIUs line in the file*/
					for (int j = 2; j < lineInfo.Length; j++)
					{
						/*Try/catch will avoid converting blank values of an AIU
							line in "aius.txt" to an index.*/
						try
						{
							int sensorID = Convert.ToInt32(lineInfo[j]);
							sensors[sensorID] = new Sensor(sensorID);
							tempSensors[j - 2] = sensors[sensorID];
						}
						catch
						{
							tempSensors[j - 2] = null;
						}
					}

					//Set the AIU object's sensors to the temporary sensors array
					AIUs[i].setSensors(tempSensors);
				}

				/*Try/catch will attempt to read initial values. If it is unable, something is wrong
					with the DCC port, and the application will close.*/
				try
				{
					CallDCC.PollAIU(AIUs[i], trains);
				}
				catch (TimeoutException)
				{
					showNotification(GLOBALS.FAILED_PORT);
					return; //This return should not occur, FAILED_PORT notification closes application
				}
			}

		}
		private void setupTracks()
		{
			/**************************************************************
			 *	Function Name:		setupTracks
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function reads from the "tracks.txt" file
			 *				to input into the tracks array all track sections
			 *				that are present in the file.
			 *				
			 *				The input parameters to the track class are:
			 *					* ID			int			0
			 *					* LEFT TOP 		Track		1
			 *					* LEFT BOTTOM	Track		2
			 *					* RIGHT TOP		Track		3
			 *					* RIGHT BOTTOM	Track		4
			 *					* Speed limit	int			5
			 *					* Turnout		Turnout		6
			 *					* LEFT Sensor	Sensor		7
			 *					* RIGHT Sensor	Sensor		8
			 *				
			 *				The track method setConnections is called to tie
			 *				the sections together if they are connected on the
			 *				railroad. The protocol for this is:
			 *					* LEFT TOP is the section in the TOP direction
			 *						of a turnout to the section's LEFT.
			 *					* LEFT BOTTOM is the section in the BOTTOM direction
			 *						of a turnout to the section's LEFT.
			 *					* RIGHT TOP is the section in the TOP direction
			 *						of a turnout to the section's RIGHT.
			 *					* RIGHT BOTTOM is the section in the BOTTOM direction
			 *						of a turnout to the section's RIGHT.
			 *				 	
			 *				Any one of these connections is allowed to be null
			 *				by being an empty string in the file. A try/catch
			 *				is used when converting the string to detect the
			 *				null locations.
			 *				
			 *				The turnout of the section is allowed to be null
			 *				by being an empty string in the file. A try/catch
			 *				is used when converting the string to detect the
			 *				null turnout.
			 *				
			 *				The speed limit of a section is set to the GLOBAL
			 *				DEFAULT_SPEED if it is not present in the file.
			 *				
			 *				The LEFT and RIGHT sensors are set to a Sensor object
			 *				in the sensors array if the value is not null in the
			 *				file.
			 * ************************************************************/

			string[] lines = File.ReadAllLines("tracks.txt");
			tracks = new Track[lines.Length];

			/*Create n track objects, where n is the number of lines in the file
				they will be unconnected tracks until they are filled in the next loop*/
			for (int i = 0; i < lines.Length; i++)
			{
				tracks[i] = new Track(i);   //Create new track object with ID of i
			}

			//Connect the track sections by reading the tracks file
			for (int i = 0; i < lines.Length; i++)
			{
				//Split the file line (comma delimmited)
				string[] lineInfo = lines[i].Split(',');

				//Create temp tracks to fill the current track connections. This step is necessary to handle null indices.
				Track[] tempTracks = new Track[4];
				for (int j = 0; j < 4; j++)
				{
					try
					{
						//The index for lineInfo is j + 1, as index 0 is the track's own ID
						tempTracks[j] = tracks[Convert.ToInt32(lineInfo[j + 1])];
					}
					catch
					{
						tempTracks[j] = null;
					}
				}

				//Set the connections for the track section
				tracks[Convert.ToInt32(lineInfo[0])].setConnections(
					tempTracks[0],      //	Left top
					tempTracks[1],      //	Left bottom
					tempTracks[2],      //	Right top
					tempTracks[3]       //	Right bottom
				);

				//Set the speed limit of the track section
				int speed;
				try
				{
					speed = Convert.ToInt32(lineInfo[5]);
				}
				catch
				{
					speed = GLOBALS.DEFAULT_SPEED;
				}
				tracks[Convert.ToInt32(lineInfo[0])].setSpeedLimit(speed);

				//Set the track's turnout
				Turnout tempTurnout;
				try
				{
					tempTurnout = turnouts[Convert.ToInt32(lineInfo[6])];
				}
				catch
				{
					tempTurnout = null;
				}
				tracks[i].setTurnout(tempTurnout);

				/*Create a temporary sensor array to hold the two sensors
					a track may have.*/
				Sensor[] tempSensors = new Sensor[2];

				//Tie the sensors to the track object
				for (int j = 0; j < 2; j++)
				{
					try
					{
						/*Get the sensor ID from the track's values. LEFT sensor is
							value 7, and the right value 8.*/
						int sensorID = Convert.ToInt32(lineInfo[7 + j]);
						tempSensors[j] = sensors[sensorID];
					}
					catch
					{
						tempSensors[j] = null;
					}
				}

				//Set the Track object's sensors to the temporary sensors array
				tracks[i].setSensors(tempSensors);
			}
		}
		private void setupTrains()
		{
			/**************************************************************
			 *	Function Name:		setupTrains
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function will initialize the train array
			 *				to MAX_TRAINS number of elements. The trains
			 *				file will then be opened to be cleared of any
			 *				remaning data from prior use.
			 *				 
			 *				This function will continue to attempt to access
			 *				the file which may be locked by another process.
			 * ************************************************************/

			trains = new Train[GLOBALS.MAX_TRAINS];
			bool opened = false;    //Flag to escape while loop

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					//Open the file and erase all contents
					FileStream fs = new FileStream("trains.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					fs.SetLength(0);
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}

		//string[] lines = File.ReadAllLines("trains.txt");
		//trains = new Train[lines.Length];

		///*Create n train objects, where n is the number of lines in the file*/
		//for (int i = 0; i < lines.Length; i++)
		//{
		//	//Split the file line (comma delimmited)
		//	string[] lineInfo = lines[i].Split(',');

		//	//Try to parse the train location. Set to null if string was null.
		//	Track tempTrack;
		//	try
		//	{
		//		tempTrack = tracks[Convert.ToInt32(lineInfo[2])];
		//	}
		//	catch (FormatException) //catches if the Track string is null
		//	{
		//		tempTrack = null;
		//	}

		//	//Create the train
		//	trains[i] = new Train(
		//		Convert.ToInt32(lineInfo[0]),                       //ID
		//		Convert.ToInt32(lineInfo[1]),                       //Address
		//		Convert.ToInt32(lineInfo[3]),                       //Speed
		//		Convert.ToInt32(lineInfo[4]),                       //Direction
		//		Convert.ToInt32(lineInfo[5]),                       //Orientation
		//		Convert.ToInt32(lineInfo[6]),                       //Status
		//		Convert.ToBoolean(Convert.ToInt32(lineInfo[7])),    //Headlight
		//		Convert.ToBoolean(Convert.ToInt32(lineInfo[8]))     //Horn
		//	);

		//	trains[i].setLocation(tempTrack);   //***SMJ*** CHECK FOR BAD DATA
		//}
	}
		private void setupTurnouts()
		{
			/**************************************************************
			 *	Function Name:		setupTurnouts
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function reads from the "turnouts.txt" file
			 *				to input into the turnout array all turnouts that
			 *				are present in the file.
			 *				
			 *				The input parameters to the turnout class are:
			 *					* ID			int
			 *					* Address		int
			 *					* Direction		int (GLOBALS)
			 *					* Occupied		bool
			 * ************************************************************/

			string[] lines = File.ReadAllLines("turnouts.txt");
			turnouts = new Turnout[lines.Length];

			/*Create n turnout objects, where n is the number of lines in the file*/
			for (int i = 0; i < lines.Length; i++)
			{
				//Split the file line (comma delimmited)
				string[] lineInfo = lines[i].Split(',');

				//Create the turnout
				turnouts[i] = new Turnout(
					Convert.ToInt32(lineInfo[0]),		//ID
					Convert.ToInt32(lineInfo[1]),		//Address
					Convert.ToInt32(lineInfo[2])		//Direction
				);

				/*Try/catch will attempt to set physical turnouts to match data structures.
					If it is unable, something is wrong with the DCC port, and the application
					will close.*/
				try
				{
					CallDCC.TurnoutDirection(turnouts[i]);
				}
				catch (TimeoutException)
				{
					showNotification(GLOBALS.FAILED_PORT);
					return; //This return should not occur, FAILED_PORT notification closes application
				}
			}
		}

		//File Read/Write Functions
		private bool readTurnouts()
		{
			/**************************************************************
			 *	Function Name:		readTurnouts
			 *	Input Parameters:	none
			 *	Output Parameters:	TRUE
			 *	
			 *	Purpose:	This function reads from the "turnouts.txt" file
			 *				to evaluate if a turnout needs to be flipped.
			 *				This function will continue to attempt to access
			 *				the file which may be locked by another process.
			 *				
			 *				The input parameters from the turnout file are:
			 *					* ID			int
			 *					* Address		int
			 *					* Direction		int
			 *					* Occupied		bool
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("turnouts.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					string[] lines = new string[turnouts.Length];

					//Loop through all lines in the file
					for (int i = 0; i < lines.Length; i++)
					{
						lines[i] = sr.ReadLine();   //Read a line
						string[] lineinfo = lines[i].Split(',');
						int id = Convert.ToInt32(lineinfo[0]);          //Get the ID of the turnout
						int direction = Convert.ToInt32(lineinfo[2]);   //Get direction of the turnout

						/*If the turnout direction in the file is different than the corresponding turnout,
							then attempt to flip the turnout*/
						if (direction != turnouts[id].Direction())
						{
							turnouts[id].Flip();
						}

						//Get the information to write back to the file
						lines[i] = turnouts[id].GetInfoLine();
					}

					//Erase file contents and open StreamWriter
					fs.SetLength(0);
					StreamWriter sw = new StreamWriter(fs);

					//Write back the lines to the file
					for (int i = 0; i < lines.Length; i++)
					{
						sw.WriteLine(lines[i]);
					}

					sw.Close();
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
			return true;
		}
		private bool readTrains()
		{
			/**************************************************************
			 *	Function Name:		readTrains
			 *	Input Parameters:	none
			 *	Output Parameters:	TRUE
			 *	
			 *	Purpose:	This function reads from the "trains.txt" file
			 *				to evaluate if a train's speed has been modified
			 *				by a CAB. This function will continue to attempt
			 *				to access the file which may be locked by another
			 *				process.
			 *				
			 *				The input parameters from the turnout file are:
			 *					* ID			int				0
			 *					* Address		int				1
			 *					* Location		Track			2
			 *					* Speed			int				3
			 *					* Direction		int (GLOBALS)	4
			 *					* Orientation	int (GLOBALS)	5
			 *					* Status		int (GLOBALS)	6
			 *					* Headlight		bool			7
			 *					* Horn			bool			8
			 *					* Occupied		bool			9
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("trains.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					string[] lines = new string[trains.Length];

					//Loop through all lines in the file
					for (int i = 0; i < lines.Length; i++)
					{
						lines[i] = sr.ReadLine();                   //Read a line

						//If there is a train on the line
						if (lines[i] != null && lines[i] != string.Empty)
						{
							string[] lineinfo = lines[i].Split(',');
							int id = Convert.ToInt32(lineinfo[0]);      //Get the ID of the train
							int speed = Convert.ToInt32(lineinfo[3]);   //Get the speed of the train

							/*Get the occupied status of the train. Set the status if it is different,
								either occupying or releasing the status of the train for use by CABs.*/
							bool occupied = Convert.ToBoolean(Convert.ToInt32(lineinfo[9]));
							if(occupied != trains[id].isOccupied())
							{
								if(occupied)
								{
									trains[id].Occupy();
								}
								else
								{
									trains[id].Release();
								}
							}

							/*If the train speed in the file is different than the corresponding train's,
								then attempt to change it*/
							if (speed != trains[id].Speed())
							{
								trains[id].setSpeed(speed);
							}

							//Get the information to write back to the file
							lines[i] = trains[id].GetInfoLine();
						}
					}

					//Erase file contents and open StreamWriter
					fs.SetLength(0);
					StreamWriter sw = new StreamWriter(fs);

					//Write back the lines to the file
					for (int i = 0; i < lines.Length; i++)
					{
						sw.WriteLine(lines[i]);
					}

					sw.Close();
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
			return true;
		}
		private bool writeTrain(Train train)
		{
			/**************************************************************
			 *	Function Name:		writeTrain
			 *	Input Parameters:	Train train
			 *	Output Parameters:	bool found
			 *	
			 *	Purpose:	This function writes to the "trains.txt" file
			 *				to output the train parameter. This function 
			 *				will continue to attempt to access the file
			 *				which may be locked by another process.
			 *				
			 *				This function will assess if the sent train is
			 *				present in the file, or if it needs to be added.
			 *				If the file already has MAX_TRAINS number of
			 *				filled lines, then this function will return
			 *				FALSE;				
			 *				
			 *				The output parameters to the train file are:
			 *					* ID			int				0
			 *					* Address		int				1
			 *					* Location		Track			2
			 *					* Speed			int				3
			 *					* Direction		int (GLOBALS)	4
			 *					* Orientation	int (GLOBALS)	5
			 *					* Status		int (GLOBALS)	6
			 *					* Headlight		bool			7
			 *					* Horn			bool			8
			 *					* Occupied		bool			9
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop
			bool found = false;		//Flag to see if the sent train is in the file

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("trains.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);
					
					string[] lines = new string[trains.Length];

					//Loop through the file to see if the train is already in the file
					for (int i = 0; i < lines.Length; i++)
					{
						lines[i] = sr.ReadLine();

						//If there is a train on the line
						if (lines[i] != null && lines[i] != string.Empty)
						{
							string[] lineinfo = lines[i].Split(',');
							int id = Convert.ToInt32(lineinfo[0]);      //Get the ID of the train

							/*If the ID of the train in the file matches the parameter, then
								set the line to the InfoLine of the train.*/
							if (id == train.ID())
							{
								lines[i] = train.GetInfoLine();
								found = true;
							}
						}
					}
					
					//Add the train to the file if it was not found in the file
					if (!found)
					{
						//Look through the trains array for an open position
						for (int i = 0; i < trains.Length; i++)
						{
							//If an open position is found
							if(trains[i] != null)
							{
								if (trains[i].ID() == train.ID())
								{
									//Fill that position in the lines array with the train's info
									lines[i] = train.GetInfoLine();
									found = true;
								}
							}
						}
					}
					
					//Erase file contents and open StreamWriter
					fs.SetLength(0);
					StreamWriter sw = new StreamWriter(fs);

					//Write back the lines to the file
					for (int i = 0; i < lines.Length; i++)
					{
						sw.WriteLine(lines[i]);
					}

					sw.Close();
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
			return found;	//False if the sent train was not found AND unable to be inserted.
		}
		private bool writeDeletedTrain(int index)
		{
			/**************************************************************
			 *	Function Name:		writeTrain
			 *	Input Parameters:	Train train
			 *	Output Parameters:	bool found
			 *	
			 *	Purpose:	This function writes to the "trains.txt" file
			 *				to delete a train with an ID that matches the
			 *				index parameter. This function  will continue
			 *				to attempt to access the file which may be
			 *				locked by another process.
			 *				
			 *				This function will assess if the sent index is
			 *				present in the file. If it is not, FALSE is
			 *				returned. Otherwise the index was found, and
			 *				TRUE is returned.	
			 *				
			 *				The output parameters to the train file are:
			 *					* ID			int				0
			 *					* Address		int				1
			 *					* Location		Track			2
			 *					* Speed			int				3
			 *					* Direction		int (GLOBALS)	4
			 *					* Orientation	int (GLOBALS)	5
			 *					* Status		int (GLOBALS)	6
			 *					* Headlight		bool			7
			 *					* Horn			bool			8
			 *					* Occupied		bool			9
			 * ************************************************************/

			bool opened = false;    //Flag to escape while loop
			bool found = false;     //Flag to see if the index was found

			//Loop until the file can be opened. Used to prevent racing conditions on file
			while (!opened)
			{
				try
				{
					FileStream fs = new FileStream("trains.txt", FileMode.Open, FileAccess.ReadWrite, FileShare.None);
					StreamReader sr = new StreamReader(fs);

					string[] lines = new string[trains.Length];

					//Loop through the file to find the train in the file
					for (int i = 0; i < lines.Length; i++)
					{
						lines[i] = sr.ReadLine();

						//If there is a train on the line
						if (lines[i] != null && lines[i] != string.Empty)
						{
							string[] lineinfo = lines[i].Split(',');
							int id = Convert.ToInt32(lineinfo[0]);      //Get the ID of the train

							/*If the ID of the train in the file matches the parameter, then
								set the line string.Empty to remove the train.*/
							if (id == index)
							{
								lines[i] = string.Empty;
								found = true;
							}
						}
					}
					
					//Erase file contents and open StreamWriter
					fs.SetLength(0);
					StreamWriter sw = new StreamWriter(fs);

					//Write back the lines to the file
					for (int i = 0; i < lines.Length; i++)
					{
						sw.WriteLine(lines[i]);
					}

					sw.Close();
					sr.Close();
					fs.Close();
					opened = true;
				}
				catch (IOException) { } //Nothing to do for exception, while loop will try again
			}
			return found;   //False if the sent train was not found AND unable to be inserted.
		}
	}
	class GLOBALS
	{
		//Directions and orientations
		public const int LEFT = 0;
		public const int RIGHT = 1;

		//Turnout states
		public const int TOP = 0;
		public const int BOTTOM = 1;

		//Track speed limit
		public const int DEFAULT_SPEED = 5;
		public const int STOPPED_SPEED = 0;

		//Train states
		public const int HALTED = 0;
		public const int DEAD_END = 1;
		public const int MOVING = 2;

		//Track states
		public const string FREE = "";
		public const string AT = "x AT";
		public const string OWNS = "x OWNS";

		//Form sizing constants
		public const int tbWidth = 40;
		public const int tbHeight = 20;
		public const int lblWidth = 55;
		public const int lblHeight = 20;
		public const int xOffset = 20;
		public const int yOffset = 10;
		public const int xBuffer = 5;
		public const int yBuffer = 15;

		//Form message constants
		public const int STOPPED = 0;
		public const int FAILED_TRAIN_ADD_ADDRESS = 1;
		public const int FAILED_TRAIN_ADD_LOCATION_INVALID = 2;
		public const int FAILED_TRAIN_ADD_LOCATION_TAKEN = 3;
		public const int FAILED_TRAIN_DELETE_OCCUPIED = 4;
		public const int FAILED_TRAIN_DELETE_INVALID = 5;
		public const int FAILED_PORT = 6;

		//Misc.
		public const int MAX_TRAINS = 5;
	}
}