﻿/************************************************************
 *	Program Name:	CCR Server.exe
 *	
 * 	Author:			Steve Vandenbush
 * 	
 * 	Modifiers:		Brice Hilgemann
 * 					Samuel Joski
 * 					
 * 	Documentation:	Samuel Joski
 * 	
 * 	Purpose:		Allows a user to select a COM port which is
 * 					used by the server program to communicate
 * 					with the CCR.
 * 					
 * 					Code originally written as PortSelect.cs
 * 					by Steve Vandenbush for communication between
 * 					RS232 serial - to - USB adapter.
 * 					
 *	Modifications:	10 March 2010 (Brice)
 *					 * Brice's adaptation of Steve's code for
 *						his project.
 *						
 *					04 April 2018 (Samuel)
 *					 * Updated form element names for clarity.
 *					 * Redocumented the functions in a more
 *						clear format.
 *					 * Modified function names for clarity.
 *					 * Adapted code to fit Spring 2018 project.
 *					 * 				
 *					
 * *********************************************************/

using System;
using System.Windows.Forms;

namespace CCRServer
{
    public partial class ComSelect : Form
    {
        public ComSelect()
        {
            InitializeComponent();
        }

        private void onComSelectShow(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		ComSelect_Shown
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is called from the ComSelect.Shown
			 *				EventHandler. It forces a refresh of the COM list
			 *				so that any available COM ports are immediately
			 *				shown.
			 * ************************************************************/
			onBtnRefreshClick(null, null);
        }
        private void onBtnRefreshKeyDown(object sender, KeyEventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onBtnRefreshKeyDown
			 *	Input Parameters:	object sender, KeyEventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is attached to the button object
			 *				btnRefresh. It is called when a key is pressed,
			 *				though the refresh of the COM ListBox only occurs
			 *				if the ENTER key was pressed.
			 *				
			 *				onBtnRefreshClick is called to refresh the COM
			 *				port ListBox if the ENTER key was pressed.
			 * ************************************************************/
			if (e.KeyCode == Keys.Enter)
            {
                onBtnRefreshClick(null, null);
            }
        }
        private void onBtnSelectPortKeyDown(object sender, KeyEventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onBtnSelectPortKeyDown
			 *	Input Parameters:	object sender, KeyEventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is attached to the button object
			 *				btnSelectPort. It is called when a key is pressed,
			 *				though the selection of the COM in the ListBox
			 *				only occurs if the ENTER key was pressed.
			 *				
			 *				onBtnSelectPortClick is called to select the COM
			 *				port in lbAvailablePorts if the ENTER key was
			 *				pressed.
			 * ************************************************************/
			if (e.KeyCode == Keys.Enter)
            {
                onBtnSelectPortClick(null, null);
            }
        }
        private void onRefreshTimerTick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onRefreshTimerTick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is attached to RefreshTimer. On
			 *				a timer tick event, onBtnRefreshClick is called
			 *				to automatically refresh the COM ports displayed
			 *				in the ListBox lbAvailablePorts.
			 * ************************************************************/
			onBtnRefreshClick(null, null);
        }
        private void onMenuStripItemDisconnectClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onMenuStripItemDisconnectClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is attached to the disconnect button
			 *				in the MenuStrip. When the button is clicked, the
			 *				COMSelect form will close.
			 * ************************************************************/
			Application.Exit();
        }
        private void onBtnRefreshClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onBtnRefreshClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is attached to the button btnRefresh.
			 *				When clicked, the ListBox lbAvailablePorts is 
			 *				cleared and repopulated from System.IO.Ports.SerialPort.
			 *				GetPortNames() function.
			 *				
			 *				If no serial ports were found, then a message
			 *				is displayed to inform the user.				
			 * ************************************************************/

			// Clear lbAvailablePorts
			lbAvailablePorts.Items.Clear();

            foreach (string PortNames in System.IO.Ports.SerialPort.GetPortNames())
            {
				// Add the ports to lbAvailablePorts
				lbAvailablePorts.Items.Add(PortNames);
            }

            if (lbAvailablePorts.Items.Count > 0)
            {
				// Select the first index if there is one
				lbAvailablePorts.SelectedIndex = 0;
            }
            else if (lbAvailablePorts.Items.Count <= 0)
            {
                MessageBox.Show("No Ports Available");
            }
        }
        private void onBtnSelectPortClick(object sender, EventArgs e)
		{
			/**************************************************************
			 *	Function Name:		onBtnSelectPortClick
			 *	Input Parameters:	object sender, EventArgs e
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function is attached to the button
			 *				btnSelectPort. When clicked, the ListBox
			 *				lbAvailablePorts is checked to see if the user 
			 *				has made a selection. If they have, the the Server
			 *				object is created and shown, and ComSelect is hid.
			 *				
			 *				If no selection was made, then a message is 
			 *				displayed to inform the user.
			 * ************************************************************/

			//If a selection was made
			if (lbAvailablePorts.SelectedIndex >= 0)
			{
				string select = lbAvailablePorts.SelectedItem.ToString();           //Store the selected COM as a string
				RefreshTimer.Enabled = false;										//Disable the RefreshTimer
				CCRServer Server = new CCRServer(select);							//Create the Server
				Hide();                                                             //Hide ComSelect
				Server.ShowDialog();                                                //Show the Server
				Show();
            }
            else
                MessageBox.Show("No Selection Made");
        }
    }
}