﻿/************************************************************
 *	File:		CommandDCC.cs
 *
 * 	Author:		Samuel Joski
 *	
 *	Contents:	Class CommandDCC
 * 	
 * 	Purpose:	This class is responsible for sending commands
 * 				to DCC through the COM port. Functions of
 * 				this class accept Train and Turnout objects
 * 				to transform their data to commands to the
 * 				railroad.
 * 					
 * 				Code originally adapted from the following
 * 				functions in Brice Hilgemann's "Form1.cs":
 * 					* InitDCC
 * 					* CommandDCC
 * 					* CheckAIU
 * 					* SetupForwardCommand
 * 					* SetupBackwardCommand
 * 					* SetupTurnoutCommand
 * 					
 * 				It is important that any Write() commands to
 * 				the COM port is followed by a Read() to a
 * 				buffer, as the command the port receives needs
 * 				to finish processing before the server can
 * 				continue.
 * *********************************************************/
using System;
using System.Collections.Generic;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CCRServer
{
	public class CommandDCC
	{
		private SerialPort port;				//The port to which commands will be written
		private const int SPEED_OFFSET = 40;	//Speed offset used, otherwise trains will be slow
		
		public CommandDCC(string com_port)
		{
			/**************************************************************
			 *	Function Name:		CommandDCC
			 *	Input Parameters:	string com_port
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function creates the SerialPort to which
			 *				commands will be written. The attributes of
			 *				the port are then set.
			 *				
			 *				This function was adapted from Brice Hilgemann's
			 *				InitDCC function in his "Form1.cs"
			 * ************************************************************/
			port = new SerialPort(com_port, 9600, Parity.None, 8, StopBits.One);
			port.BaudRate = 9600;
			port.DataBits = 8;
			port.Parity = Parity.None;
			port.StopBits = StopBits.One;
			port.DtrEnable = true;
			port.ReadTimeout = 500;
			port.Handshake = Handshake.None;
			port.DtrEnable = true;
			port.Open();
		}
		public void Disconnect()
		{
			/**************************************************************
			 *	Function Name:		Disconnect
			 *	Input Parameters:	none
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function disconnects from the SerialPort.
			 *				The port is closed to open it up for other
			 *				connections.
			 * ************************************************************/
			port.Close();
		}
		public void TrainSpeed(Train train)
		{
			/**************************************************************
			 *	Function Name:		TrainSpeed
			 *	Input Parameters:	Train train
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function accepts a Train parameter and 
			 *				converts its data to create a command to the
			 *				DCC box.
			 *				
			 *				The command is an array of 6 bytes. The meaning
			 *				of the bytes are:
			 *					[0] - Command type
			 *					[1] - High address
			 *					[2] - Low address (train address)
			 *					[3] - Direction
			 *					[4] - Speed
			 *					[5] - EMPTY
			 *					
			 *				The speed of train needs to be converted to its
			 *				absolute value, multiplied by 10, and then added
			 *				to an offset in order for the speeds to be
			 *				substantially different from one another.
			 *				
			 *				port.Write() is then used to send the command 
			 *				through the serial port to the DCC.
			 *				
			 *				port.Read() needs to be called after a write
			 *				to ensure that the communication to the track
			 *				is complete before the code continues.
			 * ************************************************************/
			byte[] command = new byte[6];
			byte[] buffer = new byte[2];

			command[0] = 0xA2;						// Standard command for motors
			command[1] = 0;							// High address, set to 0 to be simple
			command[2] = (byte) train.Address();	// Train Address

			//Determine if train is moving forwards or backwards
			if (train.Direction() == train.Orientation())
			{
				command[3] = 02;	//(byte)02 is forwards
			}
			else
			{
				command[3] = 01;	//(byte)01 is backwards
			}

			/*Use Math.abs to convert reverse speed in Train data to DCC speed.
				Also use SPEED_OFFSET to avoid miniscule speed to the physical
				train. If the train speed is 0, then do not use the offset.*/
			int speed;
			if (train.Speed() != 0)
			{
				/*Multiply the absolute value of the train speed by 8, and add
					the offset to get a proper increment for the DCC box.*/
				speed = (Math.Abs(train.Speed()) * 8) + SPEED_OFFSET;
			}
			else
			{
				speed = 0;
			}
			command[4] = (byte) speed;
			
			port.Write(command, 0, 5);
			port.Read(buffer, 0, 1);	//Read from DCC to ensure Write() is complete
		}
		public void TurnoutDirection(Turnout turnout)
		{
			/**************************************************************
			 *	Function Name:		TurnoutDirection
			 *	Input Parameters:	Turnout turnout
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function accepts a Turnout parameter and 
			 *				converts its data to create a command to the
			 *				DCC box.
			 *				
			 *				The command is an array of 6 bytes. The meaning
			 *				of the bytes are:
			 *					[0] - Command type
			 *					[1] - High address
			 *					[2] - Low address (turnout address)
			 *					[3] - Direction
			 *					[4] - ??? (set to 0)
			 *					[5] - EMPTY
			 *					
			 *				port.Write() is then used to send the command 
			 *				through the serial port to the DCC.
			 *				
			 *				port.Read() needs to be called after a write
			 *				to ensure that the communication to the track
			 *				is complete before the code continues.
			 * ************************************************************/
			byte[] command = new byte[6];
			byte[] buffer = new byte[2];

			command[0] = 0xAD;							// Standard accessory packet command
			command[1] = 0;								// High address, here set to 0 to be simple
			command[2] = (byte)turnout.Address();		// Address of turnout

			//Determine turnout direction byte
			if(turnout.Direction() == GLOBALS.TOP)
			{
				command[3] = 03;	//(byte)03 is TOP
			}
			else
			{
				command[3] = 04;	//(byte)04 is BOTTOM
			}

			command[4] = 0;

			port.Write(command, 0, 5);
			port.Read(buffer, 0, 1);    //Read from DCC to ensure Write() is complete
		}
		public Train[] PollAIU(AIU aiu, Train[] trains)
		{
			/**************************************************************
			 *	Function Name:		PollAIU
			 *	Input Parameters:	AIU aiu, Train trains (possibly null)
			 *	Output Parameters:	Train[] (possibly null)
			 *	
			 *	Purpose:	This function is called to poll all the sensors
			 *				of the AIU parameter, and to determine if any of
			 *				the sensors were triggered.
			 *				
			 *				The trains parameter is used to send along to
			 *				the EvaluateDCCPoll method of the AIU class to
			 *				determine which train hit an activated sensor.
			 *				The trains parameter could be null if this 
			 *				function was called during initialization of
			 *				AIUs and sensors.
			 *				
			 *				This function returns an array of trains IF
			 *				this function was called with a non-null
			 *				parameter Train[] trains. The array returned
			 *				is an array of trains that the poll of the 
			 *				sensors of the AIU found had triggered a sensor.
			 *				Null is returned if this function was called
			 *				during initializtion.
			 *				
			 *				The bit isolation done in this function was
			 *				taken from Brice Hilgemann's CheckAIU function
			 *				in his "Form1.cs", which in turn was adapted
			 *				from Chris Kratz.
			 * ************************************************************/

			byte[] command = new byte[4];
			int[] value = new int[15];
			byte[] buffer = new byte[2];
			byte[] buffer2 = new byte[2];

			command[0] = 0x9B;						// Standard command for AIU messages
			command[1] = (byte)aiu.Address();		// The address of the AIU being checked
			port.Write(command, 0, 2);
			
			port.Read(buffer, 0, 1);				// Reads the hi byte
			port.Read(buffer2, 0, 1);				// Reads the low byte

			for (int j = 0; j < 6; j++)
			{
				byte test = buffer[0];
				test = (byte)(test >> j);			// Isolate low bytes
				value[j + 8] = (test & 0x01);
			}

			for (int j = 0; j < 8; j++)
			{
				byte test = buffer2[0];
				test = (byte)(test >> j);			// Isolate high bytes
				value[j] = (test & 0x01);
			}
			
			/*If this is the initialization of the sensors, call the
				setSensorValues method of the AIU.*/
			if (trains == null)
			{
				aiu.setSensorValues(value);
				return null;
			}

			//Otherwise evaluate the poll results
			else
			{
				return aiu.EvaluateDCCPoll(value, trains);
			}
		}
	}
}
