﻿/************************************************************
 *	File:		RailroadResources.cs
 *
 * 	Author:		Samuel Joski
 *	
 *	Contents:	Class Train
 *				Class Track
 *				Class Turnout
 *				Class Sensor
 *				Class AIU
 * 	
 * 	Purpose:	These classes are responsible for holding all
 * 				information about railroad resources.
 * 				
 * 				All classes are tied together through their
 * 				respective class variables.
 * 				
 * 				Trains have a location (Track) and next(Track).
 * 				Tracks have a connections (Track), potentially
 * 				a turnout (Turnout), and up to two sensors
 * 				(sensors). Finally, AIUs have sensors (Sensor).
 * 				
 * 				Accessors allow for lookups of the railroad
 * 				resources, and the mutators are responsible
 * 				for allocating and deallocating resources
 * 				appropriately. 				
 * *********************************************************/
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace CCRServer
{
	public class Train
	{
		//Class data
		private int id;
		private int address;
		private Track location;
		private Track next;
		private int speed;
		private int direction;
		private int orientation;
		private int status;
		private bool headlight;
		private bool horn;
		private bool occupied;
		
		//Constructors
		public Train(int id, int address, Track location, int orientation)
		{
			//Throw exception if the location is not available for the train at creation
			if(!setLocation(location))
			{
				throw new ArgumentException("Location unavailable");
			}

			this.id = id;
			this.address = address;
			this.location = location;
			this.speed = GLOBALS.STOPPED_SPEED;
			this.direction = orientation;
			this.orientation = orientation;
			this.status = GLOBALS.HALTED;
			this.headlight = false;
			this.horn = false;
			this.occupied = false;
		}
		//Psuedo-destructor
		public bool Delete()
		{
			if (!occupied)
			{
				if (location != null)
				{
					location.Release();
					location = null;
				}
				if (next != null)
				{
					next.Release();
					next = null;
				}
				return true;
			}
			return false;
		}
		public void ForceDelete()
		{
			setSpeed(GLOBALS.STOPPED_SPEED);
			if (location != null)
			{
				location.Release();
				location = null;
			}
			if (next != null)
			{
				next.Release();
				next = null;
			}
		}

		//Accessors
		public int ID() { return id; }
		public int Address() { return address; }
		public Track Location() { return location; }
		public Track Next() { return next; }
		public int Speed() { return speed; }
		public int Direction() { return direction; }
		public int Orientation() { return orientation; }
		public int Status() { return status; }
		public bool Headlight() { return headlight; }
		public bool Horn() { return horn; }
		public bool isOccupied() { return occupied; }
		public string GetInfoLine()
		{
			return (
				id.ToString() + ',' +
				address.ToString() + ',' +
				location.ID().ToString() + ',' +
				speed.ToString() + ',' +
				direction.ToString() + ',' +
				orientation.ToString() + ',' +
				status.ToString() + ',' +
				Convert.ToInt32(headlight).ToString() + ',' +
				Convert.ToInt32(horn).ToString() + ',' +
				Convert.ToInt32(occupied).ToString()
			);
		}

		//Mutators
		public bool Occupy()
		{
			if(!occupied)
			{
				occupied = true;
				return true;
			}
			return false;
		}
		public bool Release()
		{
			if(occupied)
			{
				occupied = false;
				setSpeed(GLOBALS.STOPPED_SPEED);
				return true;
			}
			return false;
		}
		public bool setLocation(Track track)
		{
			if(track.Request())
			{
				if (track.Occupy())
				{
					location = track;
					return true;
				}
			}
			return false;
		}
		public bool setSpeed(int speed)
		{
			//Use Math.Abs to evaluate if a negative speed might exceed speed limit
			if (Math.Abs(speed) <= location.SpeedLimit())
			{
				/*If the sign of the new speed is not the same as the old speed,
					then swap the train direction.*/
				if(!isSameSign(speed))
				{
					//Swap the direction of the train
					if(direction == GLOBALS.LEFT)
					{
						direction = GLOBALS.RIGHT;
					}
					else
					{
						direction = GLOBALS.LEFT;
					}

					/*Release the request where the train was heading before its
						direction switched, if appropriate.*/
					if (next != null)
					{
						next.Derequest();
						next = null;
					}
				}
				
				//Change the speed
				this.speed = speed;

				/*If the sent speed is to have the train stop, set its status to HALTED and
					call the DCC to stop it.*/
				if (speed == GLOBALS.STOPPED_SPEED)
				{
					status = GLOBALS.HALTED;
					CCRServer.CallDCC.TrainSpeed(this);
				}

				/*Otherwise the speed request is a moving speed. Set the status to moving.*/
				else
				{
					status = GLOBALS.MOVING;
					
					//If train is moving from a stop, call Move() to get next section.
					if(next == null)
					{
						/*Only let the train move from a stop if it can successfully get a
							next section. Move will set status to HALTED or DEAD_END if the
							request was unsuccessful.*/
						if(Move())
						{
							CCRServer.CallDCC.TrainSpeed(this);
						}
					}
					else
					{
						CCRServer.CallDCC.TrainSpeed(this);
					}
				}
				
				return true;
			}
			return false;
		}
		public bool Move()
		{
			/*If the train has a next section it was towards which it was
				previously moving, check if the train can occupy it.*/
			if (next != null)
			{
				/*A train can occupy a section if another train is not there.*/
				if (!next.isOccupied())
				{
					location.Release();
					next.Occupy();
					location = next;		//Train is now in its previous next section
				}

				/*If the train cannot occupy it's next section, stop the train and update
					its status. Do not release its request for the Next section. Return.*/
				else
				{
					setSpeed(GLOBALS.STOPPED_SPEED);
					status = GLOBALS.HALTED;
					return false;
				}
			}

			/*As long as the train did not stop above, get the Next section of which 
				the train is heading in its the direction.*/
			Track tempTrack = null;
			try
			{
				tempTrack = location.getNextSection(direction);
			}
			//Catch if the train was heading toward a turnout that was not facing the train
			catch(Exception)
			{
				setSpeed(GLOBALS.STOPPED_SPEED);
				status = GLOBALS.HALTED;
				next = null;
				return false;
			}

			//Ensure that the tempTrack is not null (i.e. a dead-end)
			if(tempTrack != null)
			{
				/*Submit a request. If fulfilled, the train will own the next
					section as it moves towards it.*/
				if(tempTrack.Request())
				{
					next = tempTrack;
					status = GLOBALS.MOVING;
					return true;
				}

				/*Else if unable to request, another train is already there,
					and this train is halted*/
				else
				{
					setSpeed(GLOBALS.STOPPED_SPEED);
					status = GLOBALS.HALTED;
					next = null;
				}
			}

			//If tempTrack is null, then there is no current next track. Train becomes stalled.
			else
			{
				setSpeed(GLOBALS.STOPPED_SPEED);
				status = GLOBALS.DEAD_END;
				next = null;
			}
			return false;			
		}

		//Misc Functions
		private bool isSameSign(int speed)
		{
			return (((direction == orientation) && (speed >= 0)) || ((direction != orientation) && (speed <=0)));
		}
	}
	public class Track
	{
		//Class data
		private int id;
		private int speedLimit;
		private bool occupied;
		private bool held;
		private Turnout turnout;
		private Track[,] connections;
		private Sensor[] sensors;

		//Constructors
		public Track(int i)
		{
			id = i;
			speedLimit = GLOBALS.DEFAULT_SPEED;
			occupied = false;
			held = false;
			connections = new Track[2, 2];
		}

		//Accessors
		public int ID() { return id; }
		public int SpeedLimit() { return speedLimit; }
		public bool isOccupied() { return occupied; }
		public bool isHeld() { return held; }
		public Turnout Turnout() { return turnout; }
		public Track getNextSection(int direction)
		{
			int d;

			/*If the current location has two connections in the direction the train is going,
				then check the turnout direction to get the next section.*/
			if(connections[direction,GLOBALS.TOP] != null && connections[direction,GLOBALS.BOTTOM] != null)
			{
				d = connections[direction, GLOBALS.TOP].Turnout().Direction();
			}

			//Else there is only one option, default to GLOBALS.TOP
			else
			{
				d = GLOBALS.TOP;

				//Only check for faced away turnouts if the next section is not null
				if (connections[direction, d] != null)
				{
					//Check if the Next section is a turnout that can face away from the train
					if (connections[direction, d].Turnout() != null)
					{
						/*Get the the direction of the track that the train will be approaching from
							(i.e. if train travelling to the LEFT, it is coming from the RIGHT.*/
						int inverse_direction;
						if (direction == GLOBALS.LEFT)
						{
							inverse_direction = GLOBALS.RIGHT;
						}
						else
						{
							inverse_direction = GLOBALS.LEFT;
						}

						/*Check that there might be a turnout facing away from the train
							this IF statement is long and convoluted, but it attempts to 
							assess if the turnout is facing the train by looking beyond
							the turnout to assess if the track that the turnout is facing
							is indeed the one the train is attempting to request. If it
							is NOT (this is a not equal comparison, null is returned to
							indicated that the section is unavailable to be requested.*/
						if (connections[direction, d].connections[direction, d].connections[inverse_direction, connections[direction, d].Turnout().Direction()] != connections[direction, d])
						{
							throw (new Exception("Turnout not facing train"));
						}
					}
				}
			}

			return connections[direction, d];
		}
		public Sensor getSensor(int direction) { return sensors[direction]; }

		//Mutators
		public void setConnections(Track left1, Track left2, Track right1, Track right2)
		{
			connections[GLOBALS.LEFT, GLOBALS.TOP] = left1;
			connections[GLOBALS.LEFT, GLOBALS.BOTTOM] = left2;
			connections[GLOBALS.RIGHT, GLOBALS.TOP] = right1;
			connections[GLOBALS.RIGHT, GLOBALS.BOTTOM] = right2;
		}
		public void setSpeedLimit(int speed)
		{
			speedLimit = speed;
		}
		public bool Request()
		{
			if (!occupied && !held)
			{
				/*Check if this section is a turnout. If so, the turnout must be unoccupied.
					If the turnout is not occupied, the turnout will be occupied by the 
					train requesting this track.*/
				if (turnout != null)
				{
					if (turnout.Occupy())
					{
						held = true;
						return true;
					}
				}
				else
				{
					held = true;
					return true;
				}
			}
			return false;
		}
		public bool Derequest()
		{
			if (held)
			{
				held = false;

				//Release turnout if the section was a turnout
				if (turnout != null)
				{
					turnout.Release();
				}

				return true;
			}
			return false;
		}
		public bool Occupy()
		{
			if (!occupied && held)
			{
				occupied = true;
				return true;
			}
			return false;
		}
		public bool Release()
		{
			if (occupied)
			{
				occupied = false;
				held = false;
				
				//Release turnout if the section was a turnout
				if(turnout != null)
				{
					turnout.Release();
				}
				return true;
			}
			return false;
		}
		public void setTurnout(Turnout turnout)
		{
			this.turnout = turnout;
		}
		public void setSensors(Sensor[] sensors) { this.sensors = sensors; }
	}
	public class Turnout
	{
		//Class data
		private int id;
		private int address;
		private int direction;
		private bool occupied;

		//Constructors
		public Turnout(int id, int address, int direction)
		{
			this.id = id;
			this.address = address;
			this.direction = direction;
			this.occupied = false;
		}

		//Accessors
		public int ID() { return id; }
		public int Address() { return address; }
		public int Direction() { return direction; }
		public bool isOccupied() { return occupied; }
		public string GetInfoLine()
		{
			return (
				id.ToString() + ',' +
				address.ToString() + ',' +
				direction.ToString() + ',' +
				Convert.ToInt32(occupied).ToString()
			);
		}

		//Mutators
		public bool Occupy()
		{
			if(!occupied)
			{
				occupied = true;
				return true;
			}
			return false;
		}
		public bool Release()
		{
			if(occupied)
			{
				occupied = false;
				return true;
			}
			return false;
		}
		public bool Flip()
		{
			if(!occupied)
			{
				switch(direction)
				{
					case GLOBALS.TOP:
						direction = GLOBALS.BOTTOM;
						break;
					case GLOBALS.BOTTOM:
						direction = GLOBALS.TOP;
						break;
				}
				CCRServer.CallDCC.TurnoutDirection(this);
				return true;
			}
			return false;
		}
	}
	public class Sensor
	{
		//Class data
		private int id;
		private int status;
		
		//Constructors
		public Sensor(int id)
		{
			this.id = id;
			status = -1;
		} 

		//Accessors
		public int ID() { return id; }
		public int Status() { return status; }

		//Mutators
		public void setStatus(int status) { this.status = status; }
	}
	public class AIU
	{
		//Class data
		private int id;
		private int address;
		Sensor[] sensors;

		//Constructors
		public AIU(int id, int address)
		{
			this.id = id;
			this.address = address;
		}

		//Accessors
		public int ID() { return id; }
		public int Address() { return address; }

		//Mutators
		public void setSensors(Sensor[] sensors) { this.sensors = sensors; }
		public void setSensorValues(int[] sensorValues)
		{
			/**************************************************************
			 *	Function Name:		setSensorValues
			 *	Input Parameters:	int[] sensorValues
			 *	Output Parameters:	none
			 *	
			 *	Purpose:	This function will set all of the AIU's sensor's
			 *				values to the passed integer array sensorValues,
			 *				if the AIU has sensors.
			 *				
			 *				This function must ask if the sensor array of an
			 *				AIU is not null, and that the index of the sensor
			 *				is not null, as some AIUs have either no sensors,
			 *				or empty indices.
			 * ************************************************************/

			//Check to make sure AIU has assigned sensors
			if (sensors != null)
			{
				//Loop through all sensors
				for (int i = 0; i < sensors.Length; i++)
				{
					//Make sure the index of the sensor is not null (some AIUs have those)
					if (sensors[i] != null)
					{
						//Set the sensor value to the value read from the DCC box
						sensors[i].setStatus(sensorValues[i]);
					}
				}
			}
		}

		//Miscellaneous
		public Train[] EvaluateDCCPoll(int[] sensorValues, Train[] trains)
		{
			/**************************************************************
			 *	Function Name:		EvaluateDCCPoll
			 *	Input Parameters:	int[] sensorValues, Train[] trains
			 *	Output Parameters:	Train[] movedTrains
			 *	
			 *	Purpose:	This function accomplishes three tasks:
			 *					* Evaluate which sensor of an AIU was
			 *						triggered from the PollAIU method
			 *						of CommandDCC.
			 *					* Determine which train was on a track
			 *						whose sensor was in the same direction
			 *						the train was heading.
			 *					* If a corresponding train is found, then
			 *						call its Move() method to attempt
			 *						to allocate its next section.
			 *					* Returns a train array of any trains that
			 *						were found to have been moving toward
			 *						a sensor that triggered which belonged
			 *						to the AIU being evaluated.
			 * ************************************************************/			
			Train[] movedTrains = new Train[trains.Length];

			//If there are sensors assigned to the AIU
			if (sensors != null)
			{
				//Loop through all sensors that the AIU has
				for (int i = 0; i < sensors.Length; i++)
				{
					//If the sensor index is not null (some AIUs have null indices)
					if (sensors[i] != null)
					{
						//If the value of that sensor is different from AIU read value
						if (sensors[i].Status() != sensorValues[i])
						{
							//Check which train was expected to head to this sensor
							for (int j = 0; j < trains.Length; j++)
							{
								//If valid index
								if (trains[j] != null)
								{
									/*If the train's locations's sensor in the same direction which
										the train was heading is the same that was determined triggered
										by the PollAIU method, then attempt move the train.*/
									if (trains[j].Location().getSensor(trains[j].Direction()) == sensors[i])
									{
										trains[j].Move();
										movedTrains[j] = trains[j];
									}
								}
							}
						}
					}
				}
			}
			return movedTrains;
		}
	}
}